/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#include "nv-linux.h"
#include "nv_compiler.h"

/*
 * our global state; one per device
 */

static int num_nv_devices = 0;

nv_linux_state_t nv_linux_devices[NV_MAX_DEVICES] = { { { 0 } } };

/*
 * And one for the control device
 */

nv_linux_state_t nv_ctl_device = { { 0 } };
wait_queue_head_t nv_ctl_waitqueue;

static int nv_kern_probe(struct pci_dev *, const struct pci_device_id *);

static struct pci_device_id nv_pci_table[] = {
    { 
        .vendor      = PCI_VENDOR_ID_NVIDIA,
        .device      = PCI_ANY_ID,
        .subvendor   = PCI_ANY_ID,
        .subdevice   = PCI_ANY_ID,
        .class       = (PCI_CLASS_DISPLAY_VGA << 8),
        .class_mask  = ~0,
    },
    { }
};

MODULE_DEVICE_TABLE(pci, nv_pci_table);

static struct pci_driver nv_pci_driver = {
    .name     = "nvidia",
    .id_table = nv_pci_table,
    .probe    = nv_kern_probe,
};


/***
 *** EXPORTS to Linux Kernel
 ***/

static int __init nvidia_init_module(void)
{
    memset(nv_linux_devices, 0, sizeof(nv_linux_devices));

    if (pci_register_driver(&nv_pci_driver) < 0)
    {
        pci_unregister_driver(&nv_pci_driver); // XXX ???
        printk("NVRM: no NVIDIA graphics adapter found\n");
        return -ENODEV;
    }

    return 0;
}

static void __exit nvidia_exit_module(void)
{
    int i;

    printk("NVRM: nvidia_exit_module\n");

    for (i = 0; i < num_nv_devices; i++)
    {
        if (nv_linux_devices[i].dev)
        {
            nv_state_t *nv = NV_STATE_PTR(&nv_linux_devices[i]);
            release_mem_region(nv->bar.regs.address, nv->bar.regs.size);
        }
    }

    pci_unregister_driver(&nv_pci_driver);

}

module_init(nvidia_init_module);
module_exit(nvidia_exit_module);


/* avoid compiler warnings on UP kernels, 
 * when spinlock macros are defined away 
 */
#define NO_COMPILER_WARNINGS(nvl) \
    if (nvl == NULL) return

static void nv_lock_init_locks
( 
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    NO_COMPILER_WARNINGS(nvl);

    spin_lock_init(&nvl->rm_lock);

    sema_init(&nvl->ldata_lock, 1);
    sema_init(&nvl->at_lock, 1);
    NV_ATOMIC_SET(nvl->usage_count, 0);

    nvl->rm_lock_cpu = -1;
    nvl->rm_lock_count = 0;
}


/* find nvidia devices and set initial state */
static int
nv_kern_probe
(
    struct pci_dev *dev,
    const struct pci_device_id *id_table
)
{
    nv_state_t *nv;
    nv_linux_state_t *nvl;

    if ((dev->vendor != 0x10de) || (dev->device < 0x20) || 
        (dev->class != (PCI_CLASS_DISPLAY_VGA << 8)))
    {
        return -1;
    }

    if (num_nv_devices == NV_MAX_DEVICES)
    {
        printk("NVRM: maximum device number (%d) reached!\n", num_nv_devices);
        return -1;
    }

    /* initialize bus-dependent config state */
    nvl = &nv_linux_devices[num_nv_devices];
    nv  = NV_STATE_PTR(nvl);

    nvl->dev          = dev;
    nv->vendor_id     = dev->vendor;
    nv->device_id     = dev->device;
    nv->os_state      = (void *) nvl;
    nv->bus           = dev->bus->number;
    nv->slot          = PCI_SLOT(dev->devfn);

    nv_lock_init_locks(nv);
    

    nv->bar.regs.address = dev->resource[0].start;
    nv->bar.regs.size    = dev->resource[0].end - dev->resource[0].start + 1;

    nv->bar.fb.address   = dev->resource[1].start;
    nv->bar.fb.size      = dev->resource[1].end - dev->resource[1].start + 1;
    
    nv->interrupt_line = dev->irq;

    /* check common error condition */
    if (nv->interrupt_line == 0)
    {
        printk("NVRM: Can't find an IRQ for your NVIDIA card!  \n");
        printk("NVRM: Please check your BIOS settings.         \n");
        printk("NVRM: [Plug & Play OS   ] should be set to NO  \n");
        printk("NVRM: [Assign IRQ to VGA] should be set to YES \n");
        return -1;
    }

    /* sanity check the IO apertures */
    if ( (nv->bar.regs.address == 0) || (nv->bar.regs.size == 0) ||
         (nv->bar.fb.address   == 0) || (nv->bar.fb.size   == 0))
    {
        printk("NVRM: The IO regions for your NVIDIA card are invalid.\n");
        printk("NVRM: Your system BIOS may have misconfigured your graphics card.\n");

        if ((nv->bar.regs.address == 0) || (nv->bar.regs.size == 0))
        {
            printk("NVRM: bar0 (registers) appears to be wrong: 0x%x 0x%x\n",
                nv->bar.regs.address, nv->bar.regs.size);
        }

        if ((nv->bar.fb.address == 0) || (nv->bar.fb.size == 0))
        {
            printk("NVRM: bar1 (framebuffer) appears to be wrong: 0x%x 0x%x\n",
                nv->bar.fb.address, nv->bar.fb.size);
        }

        /* Clear out the data */
        memset(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }

    // request ownership of our bars
    // keeps other drivers from banging our registers.
    // only do this for registers, as vesafb requests our framebuffer and will
    // keep us from working properly
    if (!request_mem_region(nv->bar.regs.address, nv->bar.regs.size, "nvidia"))
    {
        printk("NVRM: pci_request_regions failed, aborting\n");

        /* Clear out the data */
        memset(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }

    // enable io, mem, and bus-mastering in pci config space
    if (pci_enable_device(dev) != 0)
    {
        printk("NVRM: pci_enable_device failed, aborting\n");

        pci_release_regions(dev);
        memset(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }
    pci_set_master(nvl->dev);

    printk("NVRM: %02x:%02x.%x %04x:%04x - 0x%08x [size=%dM]\n",
            nv->bus, nv->slot, PCI_FUNC(dev->devfn),
            nv->vendor_id, nv->device_id, nv->bar.regs.address,
            nv->bar.regs.size / (1024 * 1024));
    printk("NVRM: %02x:%02x.%x %04x:%04x - 0x%08x [size=%dM]\n",
            nv->bus, nv->slot, PCI_FUNC(dev->devfn),
            nv->vendor_id, nv->device_id, nv->bar.fb.address,
            nv->bar.fb.size / (1024 * 1024));

    num_nv_devices++;

    return 0;
}

